{$A+,B-,D+,E-,F-,G+,I-,L+,N-,O-,P-,Q-,R-,S-,T-,V+,X+,Y+}
{$M 16384,0,655360}
{
 Copyright 2004 P. Scott Harris psharris@magma.ca
 All rights reserved
 No warrantys of any kind stated or implied.

 H&L Associates
 21 Parkmount Crescent
 Ottawa, Ontario  K2H 5T3

 Unit to handle MAILSLOT functions under DOS. Owner creates mailslot
 and others can only write to it. Owner can read it and
 check for mail. AssignMail is used to produce a text file
 device driver so simple text messaging with mailslots can
 be implemented.

 Note: under Win9x multiple copies of the same message are sent and received
       for remotenames and * broadcasts, one
       for each network protocol (ipx,netbios,tcp/ip etc.) installed
       on the receiving machine. In a single protocol system
       it's not a problem, but if IPX, TCP/IP and Netbios over
       IPX are installed, you'll get 3 copies of the one message
       at the destination. Best bet is to sequentially number
       the messages and ignore the duplicates.
}
unit Hl_mail ;
interface
uses Objects, Dos ;

const
  msErrAccess  = 5 ;   { DOS file access denied }
  msErrInvalid = 1 ;   { invalid function number }
  msErrTime    = 121 ; { mailslot timeout error om read }

type
  TLongRec = record { high/low word access to 32-bit objects }
               loword : word ;
               hiword : word ;
             end ;
  TInfoRec = record { returned from MailSlotInfo }
               maxmsgsize  : word ; { max message size }
               slotsize    : word ; { max buffer size }
               nxtmsgsize  : word ; { size of next message if any }
               nxtmsgpri   : word ; { priority of next message if one }
               msgswaiting : word ; { number of waiting messages }
             end ;

  TParmRec = record { record used in call to WriteMailSlot }
               timeo  : longint; { timeout in msec }
               bufptr : pointer; { pointer to message to be sent }
             end ;

  TPeekRec = record { returned from PeekMailSlot and ReadMailSlot }
               bytesread : word ; { number of bytes read }
               nxtsize   : word ; { size of next pending message }
               nxtpri    : word ; { priority of next pending message }
             end ;

  PMailBuf = ^TMailBuf ; { mail slot buffer }
  TMailBuf = array[ 0..8191 ] of byte ;

  PMailSlot = ^TMailSlot ; { mail slot object }
  TMailSlot = object( TObject )
                mailslot : string ;    { name of mailslot }
                buffer   : TMailBuf ;  { I/O buffer }
                handle   : word ;      { handle returned by slot creation }
                dosregs  : registers ; { for MsDos calls }
                infodata : TInfoRec ;  { data returned by Info }
                peekdata : TPeekrec ;  { data returned by Read/Peek }
                Constructor Init( mslot : string ; msgsize : word ) ;
                Function Info : word;
                Function Read( var message ; var msgsize : word ;
                               tmo : longint ) : word ;
                Function Write( var message ; msgsize : word ;
                                tmo : longint ) : word;
                Function Peek( var message ; var msgsize : word ) : word;
                Destructor Done ; virtual ;
                private
                  Function Make( msgsize : word ) : word;
                  Function Delete : word ;
              end ;
{ Write a message to a specified mailslot anywhere on the network }
Function WriteMailSlot( mailslot : string ;
                        var message ; msgsize : word ;
                        tmo : longint) : word;

{ Return a valid mailslot name from remote and local names }
Function Slotname( remote,local : string ) : string ;

{ Assign a file to be a mailslot text device }
Procedure AssignMail( rname,lname : pathstr ; rto,wto : longint ;
                      var f : text ) ;

implementation

type
  TUserRec = record { map of user data area 16-bytes in TextRec }
               slotptr : PMailSlot ; { mailslot object for read }
               wtmo    : longint ; { write timeout }
               rtmo    : longint ; { read timeout }
               filler  : pointer ; { unused }
             end ;

{-----------------------------------------------------------
 Convert null-terminated pchar to string
-----------------------------------------------------------}
Function StrPas(Str: PChar): String;
var
  stemp : string ;
  indx  : integer ;
begin
  indx := 0 ;
  stemp := '' ;
  while (str[indx]<>#0) and (indx<Sizeof( string )) do
    begin
      stemp := stemp+str[indx] ;
      Inc( indx ) ;
    end ;
  strpas := stemp ;
end ;

{-----------------------------------------------------------
 Create a valid mail slot name given a local mail slot name
 and an optional remote network name
 local mailslot         \mailslot\localname
 remote mailslot        \\remotename\mailslot\localname
 domain                 \\domainname\mailslot\localname
 broadcast              \\*\mailslot\localname
-----------------------------------------------------------}
Function Slotname( remote,local : string ) : string ;
begin
  local := '\mailslot\' + local + #0;
  if (remote<>'') then
    local := '\\' + remote + local ;
  slotname := local ;
end ;

{-----------------------------------------------------------
Generic write to mailslot, even one not created by this app
DOS only supports Class 2 mail, priority 0

INT 21 u - LAN Manager Enhanced DOS - DosWriteMailslot
        AX = 5F52h
        BX = class
        CX = length of buffer
        DX = priority
        ES:DI -> DosWriteMailslot parameter structure (see #1066)
        DS:SI -> mailslot name
Return: CF clear if successful
        CF set on error
            AX = error code
SeeAlso: AX=5F4Fh,AX=5F50h,AX=5F51h

#1066 Format of LAN Manager DosWriteMailslot parameter structure:
Format of LAN Manager DosWriteMailslot parameter structure:

Offset  Size    Description     (Table 1066)
 00h    DWORD   timeout
 04h    DWORD   -> buffer

mailslot        name of mailslot
message         address of buffer holding message
msgsize         size in bytes of message
tmo             write delay till timeout in msec
-----------------------------------------------------------}
Function WriteMailSlot( mailslot : string ;
                        var message ; msgsize : word ;
                        tmo : longint  ) : word ;
var
  parameters : TParmRec ;
  dosregs    : registers ;

begin
  writemailslot := 0 ;
  if ( msgsize=0 ) then
    Exit ;
  Parameters.timeo  := tmo ;
  Parameters.bufptr := @message ;
  Fillchar( dosregs,Sizeof( dosregs ),0);
  with dosregs do
    begin
      AX := $5F52;
      BX := 2;                 { Class is fixed at 2 }
      CX := msgsize ;
      DX := 0;                 { Priority is fixed at 0 }
      ES := seg(Parameters);
      DI := ofs(Parameters);
      DS := seg(Mailslot[1]);
      SI := ofs(Mailslot[1]);
      MsDos( dosregs );
      if ( Flags and fCarry )<>0 then
        writemailslot := ax
    end
end ;

{-----------------------------------------------------------
 Create a mailslot with an optional network name and a
 local name. Fail if can't make the slot.

 mailslot        name of mailslot
 msgsize         max size in bytes of any message
-----------------------------------------------------------}
Constructor TMailSlot.Init( mslot : string ; msgsize : word ) ;
begin
  inherited Init ;
  mailslot := mslot ;
  Fillchar( buffer,Sizeof( buffer ),0 ) ;
  Fillchar( infodata,Sizeof( infodata ),0 ) ;
  Fillchar( peekdata,Sizeof( peekdata ),0 ) ;
  handle := Make( msgsize ) ;
  if ( Hi( handle )<>0 ) then
    Fail ;
end ;

{-----------------------------------------------------------
INT 21 u - LAN Manager Enhanced DOS - DosMakeMailslot
        AX = 5F4Dh
        BX = message size
        CX = mailslot size (must be bigger than message size by at least 1)
                           (minimum 1000h, maximum FFF6h)
                           (buffer must be 9 bytes bigger than this)
        DS:SI -> name
        ES:DI -> memory buffer
Return: CF clear if successful
            AX = handle
        CF set on error
            AX = error code
SeeAlso: AX=5F4Eh,AX=5F4Fh,AX=5F50h,AX=5F51h

 msgsize         max size in bytes of any message
-----------------------------------------------------------}
Function TMailSlot.Make( msgsize : word ) : word ;
begin
  Fillchar( dosregs,Sizeof( dosregs ),0);
  with dosregs do
    begin
      ax := $5F4D ;
      bx := msgsize ;
      cx := Sizeof( TMailBuf )-9 ;
      ES := Seg( buffer ) ;
      DI := Ofs( buffer ) ;
      DS := seg(mailslot[1]);
      SI := ofs(mailslot[1]);
      MsDos( dosregs );
      if ( Flags and fCarry )<>0 then
        ax := ax or $8000 ;
      make := ax ;
    end
end ;

{-----------------------------------------------------------
INT 21 u - LAN Manager Enhanced DOS - DosDeleteMailslot
        AX = 5F4Eh
        BX = handle
Return: CF clear if successful
            ES:DI -> memory to be freed (allocated during DosMakeMailslot)
        CF set on error
            AX = error code
SeeAlso: AX=5F4Dh,AX=5F4Fh
-----------------------------------------------------------}
Function TMailSlot.Delete : word ;
var
  bufptr : pointer ;
begin
  Fillchar( dosregs,Sizeof( dosregs ),0);
  with dosregs do
    begin
      ax := $5F4E ;
      bx := handle ;
      MsDos( dosregs );
      if ( Flags and fCarry )<>0 then
        delete := ax
      else
        begin
          bufptr := Ptr( es,di ) ;
          delete := 0 ;
        end ;
    end
end ;

{-----------------------------------------------------------
INT 21 u - LAN Manager Enhanced DOS - DosMailslotInfo
        AX = 5F4Fh
        BX = handle
Return: CF clear if successful
            AX = max message size
            BX = mailslot size
            CX = next message size
            DX = next message priority
            SI = number of messages waiting
        CF set on error
            AX = error code
SeeAlso: AX=5F4Dh,AX=5F4Eh,AX=5F50h
-----------------------------------------------------------}
Function TMailSlot.Info : word ;
begin
  Fillchar( dosregs,Sizeof( dosregs ),0);
  Fillchar( infodata,Sizeof( infodata ),0 ) ;
  with dosregs do
    begin
      ax := $5F4F ;
      bx := handle ;
      MsDos( dosregs );
      if ( Flags and fCarry )<>0 then
        info := ax
      else
        with infodata do
          begin
            maxmsgsize  := ax ;
            slotsize    := bx ;
            nxtmsgsize  := cx ;
            nxtmsgpri   := dx ;
            msgswaiting := si ;
            info := 0 ;
          end ;
    end
end ;

{-----------------------------------------------------------
INT 21 u - LAN Manager Enhanced DOS - DosPeekMailslot
        AX = 5F51h
        BX = handle
        ES:DI -> buffer
Return: CF clear if successful
            AX = bytes read
            CX = next item's size
            DX = next item's priority
        CF set on error
            AX = error code
SeeAlso: AX=5F35h,AX=5F4Fh,AX=5F50h,AX=5F52h
-----------------------------------------------------------}
Function TMailSlot.Peek( var message ; var msgsize : word ) : word ;
begin
  Fillchar( dosregs,Sizeof( dosregs ),0);
  Fillchar( peekdata,Sizeof( peekdata ),0 ) ;
  msgsize := 0 ;
  with dosregs do
    begin
      ax := $5F51 ;
      bx := handle ;
      es := TLongRec( message ).hiword ;
      di := TLongRec( message ).loword ;
      MsDos( dosregs ) ;
      if ( Flags and fCarry )<>0 then
        peek := ax
      else
        with peekdata do
          begin
            bytesread := ax ;
            msgsize   := ax ;
            nxtsize   := cx ;
            nxtpri    := dx ;
            peek      := 0 ;
          end ;
    end
end ;

{-----------------------------------------------------------
INT 21 u - LAN Manager Enhanced DOS - DosReadMailslot
        AX = 5F50h
        BX = handle
        DX:CX = timeout
        ES:DI -> buffer
Return: CF clear if successful
            AX = bytes read
            CX = next item's size
            DX = next item's priority
        CF set on error
            AX = error code
SeeAlso: AX=5F4Dh,AX=5F4Fh,AX=5F51h,AX=5F52h
-----------------------------------------------------------}
Function TMailSlot.Read( var message ; var msgsize : word ;
                         tmo : longint ) : word ;
begin
  Fillchar( dosregs,Sizeof( dosregs ),0);
  Fillchar( peekdata,Sizeof( peekdata ),0 ) ;
  msgsize := 0 ;
  with dosregs do
    begin
      ax := $5F50 ;
      bx := handle ;
      cx := TLongRec( tmo ).loword ;
      dx := TLongRec( tmo ).hiword ;
      es := Seg( message ) ;
      di := Ofs( message ) ;
      MsDos( dosregs ) ;
      if ( Flags and fCarry )<>0 then
        read := ax
      else
        with peekdata do
          begin
            bytesread := ax ;
            msgsize   := ax ;
            nxtsize   := cx ;
            nxtpri    := dx ;
            read      := 0 ;
          end ;
    end
end ;

{-----------------------------------------------------------
 Write to mailslot created by this app (not normally done)

INT 21 u - LAN Manager Enhanced DOS - DosWriteMailslot
        AX = 5F52h
        BX = class
        CX = length of buffer
        DX = priority
        ES:DI -> DosWriteMailslot parameter structure (see #1066)
        DS:SI -> mailslot name
Return: CF clear if successful
        CF set on error
            AX = error code
SeeAlso: AX=5F4Fh,AX=5F50h,AX=5F51h

#1066 Format of LAN Manager DosWriteMailslot parameter structure:
Format of LAN Manager DosWriteMailslot parameter structure:

Offset  Size    Description     (Table 1066)
 00h    DWORD   timeout
 04h    DWORD   -> buffer
-----------------------------------------------------------}
Function TMailSlot.Write( var message ; msgsize : word ;
                          tmo : longint ) : word ;
var
  parameters : TParmRec ;

begin
  Parameters.timeo  := tmo ;
  Parameters.bufptr := @message ;
  fillchar( dosregs,Sizeof( dosregs ),0);
  with dosregs do
    begin
      AX := $5F52;
      BX := 2;                 { Class is fixed at 2 }
      CX := msgsize ;
      DX := 0;                 { Priority is fixed at 0 }
      ES := seg(Parameters);
      DI := ofs(Parameters);
      DS := seg(Mailslot[1]);
      SI := ofs(Mailslot[1]);
      MsDos( dosregs );
      write := 0 ;
      if ( Flags and fCarry )<>0 then
        write := ax
    end
end ;

{-----------------------------------------------------------
 Dispose of mailslot object and return memory to general use
-----------------------------------------------------------}
Destructor TMailSlot.Done ;
begin
  Delete ;
  inherited Done ;
end ;

{$F+}
{-----------------------------------------------------------
 Open a text mailslot device. A mailbox is really only a
 read-only device that some remote machine can write to.
 Writing to our own mailbox is ignored and writing to a
 remote mailbox doesn't require any local setup.
 The mailslot device handles up to 255-byte strings.
-----------------------------------------------------------}
Function MailOpen( var f : text ) : integer ;
var
  mailslot : string ;
  mailptr  : PMailSlot ;
begin
  mailopen := 0 ;
  with TextRec( f ) do
    begin
      case mode of
        fmInput : begin { Reset : open local mailbox for read }
                    mailslot := StrPas( name ) ;
                    New( mailptr,Init( mailslot,Sizeof( string ) ) ) ;
                    TUserRec( userdata ).slotptr := mailptr ;
                    if ( mailptr=NIL ) then
                      mailopen := msErrAccess
                  end ;
       fmOutput : begin { Rewrite : open for write remote mailbox }
                  end ;
        fmInOut : begin { Append : same as open for write remote mailbox }
                    mode := fmOutPut ;
                  end ;
      else
        mailopen := msErrinvalid ; { invalid function number }
      end ;
    end ;
end ;

{-----------------------------------------------------------
 Handle I/O for text mailslot device
-----------------------------------------------------------}
Function MailInOut( var f : text ) : integer ;
var
  mailslot : string ;
  mailptr  : PMailSlot ;
  ecode    : integer ;
begin
  mailinout := 0 ;
  with TextRec( f ) do
    begin
      case mode of
        fmInput : begin { read bufend characters into bufptr^ }
                    mailptr := TUserRec( userdata ).slotptr ;
                    if ( mailptr<>NIL )then
                      with mailptr^ do
                        begin
                          mailinout := Read( bufptr^,bufend,
                                         TUserRec( userdata ).rtmo ) ;
                          bufpos := 0 ;
                        end
                    else
                      mailinout := msErrAccess ;
                  end ;
       fmOutput : begin { write bufpos characters from bufptr^ }
                    mailslot := StrPas( name ) ;
                    mailinout := WriteMailSlot( name,bufptr^,bufpos,
                                                TUserRec( userdata ).wtmo ) ;
                    bufpos := 0 ;
                  end ;
      else
        mailinout := msErrInvalid ; { invalid function number }
      end ;
    end ;
end ;

{-----------------------------------------------------------
 Flush mailslot text device
-----------------------------------------------------------}
Function MailFlush( var f : text ) : integer ;
begin
  mailflush := 0 ;
  with TextRec( f ) do
    begin
      case mode of
        fmInput : begin { seldom used }
                    bufpos := 0 ;
                    bufend := 0 ;
                  end ;
       fmOutput : begin { same as InOut }
                    mailflush := MailInOut( f ) ;
                  end ;
      else
        mailflush := msErrInvalid ; { invalid function number }
      end ;
    end ;
end ;

{-----------------------------------------------------------
 Close mailslot text device
-----------------------------------------------------------}
Function MailClose( var f : text ) : integer ;
var
  mailptr : PMailSlot ;

begin
  mailclose := 0 ;
  with TextRec( f ) do
    begin
      case mode of
        fmOutPut : mailclose := MailInout( f ) ;
      end ;
      mailptr := TUserRec( userdata ).slotptr ;
      if ( mailptr<>NIL ) then
        begin
          Dispose( mailptr,Done ) ;
          TUserRec( userdata ).slotptr := NIL ;
        end ;
    end ;
end ;

{-----------------------------------------------------------
 For text device driver assignment
 rname : remote mailslot name
 lname : local mailslot name
 rto   : read timeout value in msec
 wto   : write timeout value in msec
 f     : text file
-----------------------------------------------------------}
Procedure AssignMail( rname,lname : pathstr ; rto,wto : longint ;
                      var f : text ) ;
var
  mailslot : pathstr ;
begin
  with TextRec( f ) do
    begin
      mode := fmClosed ;
      Bufsize := Sizeof( buffer ) ;
      bufptr := @buffer ;
      openfunc := @MailOpen ;
      inoutfunc := @MailInOut ;
      flushfunc := @MailFlush ;
      CloseFunc := @MailClose ;
      Fillchar( userdata,Sizeof( userdata ),0 ) ;
      mailslot := SlotName( rname,lname ) ;
      Move( mailslot[1],name,Length( mailslot ) ) ;
{
 The 16-byte user data area is used to store a pointer to
 an instance of the mailslot object, as well as timeout
 values for reading and writing.
}
      with TUserRec( userdata ) do
        begin
          rtmo := rto ;
          wtmo := wto ;
        end ;
    end ;
end ;
{$F-}
end .